//
// WebServlet.java - Web stuff
//

//...simports:0:
import java.net.*;
import java.io.*;
import java.util.*;
import jakarta.servlet.*;
import jakarta.servlet.http.*;
import java.util.zip.*;
//...e

public class WebServlet extends HttpServlet
  {
  protected Properties props;
//...sgetInitParameter:2:
protected String getInitParameter(String p)
  {
  if ( props != null )
    {
    String v;
    if ( (v = props.getProperty(p)) != null )
      return v;
    }
  ServletConfig sc = getServletConfig();
  return sc.getInitParameter(p);
  }
//...e
//...sgetInitParameter:2:
protected String getInitParameter(String p, String def)
  {
  String v;
  if ( (v = getInitParameter(p)) != null )
    return v;
  else
    return def;
  }
//...e
//...sgetInitParameterVector:2:
protected Vector<String> getInitParameterVector(String p)
  {
  String v;
  Vector<String> vec = new Vector<String>();
  if ( (v = getInitParameter(p)) != null )
    {
    StringTokenizer st = new StringTokenizer(v, ",");
    while ( st.hasMoreTokens() )
      vec.add(st.nextToken());
    }
  return vec;
  }
//...e
//...sclass Location:2:
// Keep per-location information

protected class Location
  {
  String uri;
  String dir;
  String include;
  String exclude;
  String index;
  protected String getIP(String loc, String prop, String def)
    {
    String ip;
    if ( (ip = getInitParameter("locations."+loc+"."+prop)) != null )
      return ip;
    else if ( (ip = getInitParameter("locations.default."+prop)) != null )
      return ip;
    else
      return def;
    }
  public Location(String loc)
    {
    uri     = getIP(loc, "uri"    , "/");
    dir     = getIP(loc, "dir"    , "/");
    include = getIP(loc, "include", null);
    exclude = getIP(loc, "exclude", null);
    index   = getIP(loc, "index"  , "index.html");
    if ( uri.endsWith("/") )
      uri = uri.substring(0, uri.length()-1);
    if ( dir.endsWith("/") )
      dir = dir.substring(0, dir.length()-1);
    }
  }
//...e
  protected Vector<Location> locations = new Vector<Location>();
    // ordered list of Locations
  protected String serverName;
  protected String logfile;
  protected Dictionary<String,String> icons = new Hashtable<String,String>();
//...slog:2:
protected void log(String message)
  {
  if ( logfile != null )
    try
      {
      FileOutputStream fos = new FileOutputStream(logfile, true);
      PrintWriter pw = new PrintWriter(fos);
      pw.println(message);
      pw.flush();
      fos.close();
      }  
    catch ( FileNotFoundException fnfe )
      {
      }
    catch ( IOException ioe )
      {
      }
  }
//...e
//...ssendHeader:2:
protected void sendHeader(HttpServletResponse resp, String type)
  throws IOException
  {
  resp.setContentType("text/html");
  ServletOutputStream os = resp.getOutputStream();
  String title = serverName;
  if ( type != null )
    title += ( " - " + type );
  os.println("<HTML>");
  os.println("<HEAD>");
  os.println("<TITLE>"+title+"</TITLE>");
  os.println("</HEAD>");
  os.println("<BODY BGCOLOR=\"#FFFFFF\">");
  os.println("<H1>"+title+"</H1>");
  os.println();
  }
//...e
//...ssendTrailer:2:
protected void sendTrailer(HttpServletResponse resp)
  throws IOException
  {
  ServletOutputStream os = resp.getOutputStream();
  os.println();
  os.println("</BODY>");
  os.println("</HTML>");
  }
//...e
//...ssendFileBytes:2:
protected void sendFileBytes(File f, OutputStream os)
  throws IOException
  {
  FileInputStream fis = new FileInputStream(f);
  byte[] buf = new byte[0x10000];
  int n;
  while ( (n = fis.read(buf)) > 0 )
    os.write(buf, 0, n);
  fis.close();
  }
//...e
//...ssendFile:2:
//...sgetMimeType:2:
// If the file has a file extension, try to determine MIME type by
//   looking to see if we've got a mime.ext property
//   looking to see if the the container can map the filename to a MIME type

public String getMimeType(String fn)
  {
  String bfn = fn.substring(fn.lastIndexOf('/')+1);
  int dot;
  if ( (dot = bfn.lastIndexOf('.')) != -1 )
    {
    String ext = bfn.substring(dot+1);
    String type;
    if ( (type = props.getProperty("mime."+ext)) != null )
      return type;
    ServletContext sc = getServletConfig().getServletContext();
    if ( (type = sc.getMimeType(fn)) != null )
      return type;
    }
  return null;
  }
//...e

protected void sendFile(HttpServletResponse resp, String uri, File f)
  throws IOException
  {
  if ( ! f.canRead() )
    resp.sendError(HttpServletResponse.SC_NOT_FOUND);
  else
    {
    log("send "+uri);
    String mime;
    if ( (mime = getMimeType(f.getName())) == null )
      mime = "text/plain";
    resp.setContentType(mime);
    long len = f.length();
    if ( len <= 0x7fffffffL )
      // Help the client know how to display download progress bar
      resp.setContentLength((int) len);
    ServletOutputStream os = resp.getOutputStream();
    sendFileBytes(f, os);
    }
  }
//...e
//...ssendDirectory:2:
//...sgetIcon:2:
// If the file has a file extension, try to determine icon
//   looking to see if the the container can map the filename to a MIME type

public String getIcon(String fn)
  {
  String bfn = fn.substring(fn.lastIndexOf('/')+1);
  int dot;
  if ( (dot = bfn.lastIndexOf('.')) != -1 )
    {
    String ext = bfn.substring(dot+1);
    String icon;
    if ( (icon = getInitParameter("icon."+ext)) != null )
      return icon;
    if ( (icon = icons.get(ext)) != null )
      return "/webserver/"+icon;
    }
  return null;
  }
//...e
//...sencodeHTML:2:
// For encoding data within HTML output.
// This doesn't try to be clever.
// It could try harder to map to HTML entities with names, such as &cent;
// It just does the common ones, and falls back to numeric for the rest.

protected String encodeHTML(String s)
  {
  StringBuilder sb = new StringBuilder( s.length() );
  for ( int i = 0; i < s.length(); i++ )
    {
    char c = s.charAt(i);
    switch ( c )
      {
      case '<': sb.append("&lt;"); break;
      case '>': sb.append("&gt;"); break;
      case '&': sb.append("&amp;"); break;
      case '"': sb.append("&quot;"); break;
      default:
        if ( c >= ' ' && c <= '~' )
          // regular printable ASCII character
          sb.append(c);
        else
          // use numeric entity notation for the rest
          {
          sb.append("&#");
          sb.append(c);
          sb.append(";");
          }
        break;
      }
    }
  return sb.toString();
  }
//...e

protected void sendDirectory(HttpServletResponse resp, String uri, File dir)
  throws IOException
  {
  log("send "+uri+"/");
  sendHeader(resp, encodeHTML(uri+"/"));
  ServletOutputStream os = resp.getOutputStream();
  os.println("<P>");
  os.println("<TABLE>");
  os.println("<TR><TD BGCOLOR=\"#CCCCFF\">Icon<TD BGCOLOR=\"#CCCCFF\">Name<TD BGCOLOR=\"#CCCCFF\">Size");
  String bg;
  bg = " BGCOLOR=\"#FFFFFF\"";
  os.print("<TR><TD"+bg+"><IMG SRC=\"/webserver/up.gif\" ALT=\"\"><TD"+bg+"><A HREF=\"..\">..</A><TD"+bg+" ALIGN=RIGHT>-");
  String[] fs = dir.list();
  Arrays.sort(fs);
  for ( int i = 0; i < fs.length; i++ )
    {
    if ( (i & 1) == 0 )
      bg = " BGCOLOR=\"#EEEEEE\"";
    else
      bg = " BGCOLOR=\"#FFFFFF\"";
    String fn = fs[i];
    String fnH = encodeHTML(fn);
    File f = new File(dir.getCanonicalPath()+"/"+fn);
    if ( f.isDirectory() )
      os.println("<TR><TD"+bg+"><IMG SRC=\"/webserver/directory.gif\" ALT=\"\">"+
                     "<TD"+bg+"><A HREF=\""+resp.encodeURL(fnH+"/")+"\">"+fnH+"</A>"+
                             " (<A HREF=\""+resp.encodeURL(fnH+".zip")+"?mode=zipdir\">.zip</A>)"+
                     "<TD"+bg+" ALIGN=RIGHT>-");
    else
      {
      String icon;
      if ( (icon = getIcon(fn)) == null )
        icon = "/webserver/unknown.gif";
      os.println("<TR><TD"+bg+"><IMG SRC=\""+icon+"\" ALT=\"\">"+
                     "<TD"+bg+"><A HREF=\""+resp.encodeURL(fnH)+"\">"+fnH+"</A>"+
                     "<TD"+bg+" ALIGN=RIGHT>"+Long.toString(f.length()));
      }
    }  
  os.println("</TABLE>");
  sendTrailer(resp);
  }
//...e
//...ssendDirectoryAsZip:2:
//...ssendDirectoryToZipStream:2:
protected void sendDirectoryToZipStream(
  ZipOutputStream zos,
  String prefix, File dir
  )
  throws IOException
  {
  // Store a ZipEntry corresponding to the directory
  // Doesn't seem well documented as to how this is actually done.
  // By looking at the source for an implementation of ZipEntry,
  // I see that the trailing / indicates a directory.
  ZipEntry ze = new ZipEntry(prefix);
  zos.putNextEntry(ze);
  zos.closeEntry();
  if ( dir.canRead() )
    {
    String[] fs = dir.list();
    Arrays.sort(fs);
    for ( int i = 0; i < fs.length; i++ )
      {
      String fn = fs[i];
      File f = new File(dir.getCanonicalPath()+"/"+fn);
      if ( f.isDirectory() )
        sendDirectoryToZipStream(zos, prefix+fn+"/", f);
      else if ( f.canRead() )
        {
        ZipEntry ze2 = new ZipEntry(prefix+fn);
        zos.putNextEntry(ze2);
        sendFileBytes(f, zos);
        zos.closeEntry();
        }
      }
    }
  }
//...e

protected void sendDirectoryAsZip(
  HttpServletResponse resp,
  String uri, String bfn, File dir
  )
  throws IOException
  {
  log("send "+uri+" (generated)");
  resp.setContentType("application/x-zip-compressed");
  ServletOutputStream os = resp.getOutputStream();
  ZipOutputStream zos = new ZipOutputStream(os);
  zos.setMethod(ZipEntry.DEFLATED);
  zos.setLevel(Deflater.DEFAULT_COMPRESSION);
  sendDirectoryToZipStream(zos, bfn+"/", dir);
  zos.finish();
  }
//...e
//...ssend:2:
// On input :-
//   l.uri will be "" for the root uri
//         or one or more of "/pathelem" for other uris
//   l.dir could be "" for the root directory
//         or one or more of "/pathelem" for other directories

//...sresolve:2:
// fn doesn't end in "/".
// fn has zero or more path elements, each starting with "/".

protected static String resolve(String fn)
  {
  StringTokenizer st = new StringTokenizer(fn, "/\\");
  Vector<String> v = new Vector<String>();
  while ( st.hasMoreTokens() )
    {
    String e = st.nextToken();
    if ( e.equals(".") )
      ;
    else if ( e.equals("..") )
      {
      int s = v.size();
      if ( s == 0 )
        return null;
      v.setSize(s-1);
      }
    else
      v.add(e);
    }
  fn = "";
  for ( int i = 0; i < v.size(); i++ )
    fn += ( "/" + v.elementAt(i) );
  return fn;
  }
//...e

protected void send(
  HttpServletRequest req,
  HttpServletResponse resp,
  Location l, String uri
  )
  throws IOException
  {
  String mode;
  if ( (mode = req.getParameter("mode")) == null )
    mode = "";
  if ( uri.endsWith("/") )
//...sdirectory listing:6:
{
uri = uri.substring(0,uri.length()-1);
String dir = resolve(uri.substring(l.uri.length()));
if ( dir != null )
  {
  dir = l.dir+dir;
  File f = new File(dir+"/");
    // Note that if we omit the "/" suffix when we are web serving the
    // root directory (rare I'll admit), we'll end up invoking File("")
    // which doesn't work as expected.
  if ( f.isDirectory() )
    {
    if ( ! l.index.equals("") )
      {
      File fi = new File(dir+"/"+l.index);
      if ( fi.exists() )
        {
        sendFile(resp, uri, fi);
        return;
        }
      }
    sendDirectory(resp, uri, f);
    return;
    }
  }
}
//...e
  else
//...sreturn file data:6:
{
String fn = resolve(uri.substring(l.uri.length()));
if ( fn != null )
  {
  fn = l.dir+fn;
  if ( fn.equals("") )
    fn = "/";
  if ( mode.equals("zipdir") )
    {
    if ( fn.endsWith(".zip") )
      {
      String dir = fn.substring(0,fn.length()-4);
      File f2 = new File(dir+"/");
      if ( f2.isDirectory() )
        {
        String bfn = dir.substring(dir.lastIndexOf('/')+1);
        sendDirectoryAsZip(resp, uri, bfn, f2);
        return;
        }
      }
    }
  else
    {  
    File f = new File(fn);
    if ( f.isFile() )
      {
      sendFile(resp, uri, f);
      return;
      }
    if ( f.isDirectory() )
      {
      resp.sendRedirect(resp.encodeRedirectURL(uri+"/"));
      return;
      }
    }
  }
}
//...e
  resp.sendError(HttpServletResponse.SC_NOT_FOUND);
  }
//...e
//...sdoGet:2:
protected void doGet(HttpServletRequest req, HttpServletResponse resp)
  throws ServletException, IOException
  {
  String uri = URLDecoder.decode( req.getRequestURI(), "UTF-8" );
  for ( int i = 0; i < locations.size(); i++ )
    {
    Location l = locations.elementAt(i);
    if ( uri.equals(l.uri) ||
         uri.startsWith(l.uri+"/") )
      {
      send(req, resp, l, uri);
      return;
      }
    }
  resp.sendError(HttpServletResponse.SC_NOT_FOUND);
  }
//...e
//...ssetupIcons:2:
// A selection of mappings from file extension to icons
// Can be overridden or extended in properties

protected void setupIcons()
  {
  icons.put("doc",    "word.gif");
  icons.put("xls",    "excel.gif");
  icons.put("ppt",    "powerpoint.gif");
  icons.put("mpp",    "project.gif");
  icons.put("docx",   "word.gif");
  icons.put("xlsx",   "excel.gif");
  icons.put("pptx",   "powerpoint.gif");
  icons.put("pdf",    "acrobat.gif");
  icons.put("ps" ,    "postscript.gif");
  icons.put("zip",    "archive.gif");
  icons.put("tar",    "archive.gif");
  icons.put("au",     "audio.gif");
  icons.put("snd",    "audio.gif");
  icons.put("aif",    "audio.gif");
  icons.put("aiff",   "audio.gif");
  icons.put("aifc",   "audio.gif");
  icons.put("wav",    "audio.gif");
  icons.put("gif",    "image.gif");
  icons.put("png",    "image.gif");
  icons.put("jpg",    "image.gif");
  icons.put("jpe",    "image.gif");
  icons.put("jpeg",   "image.gif");
  icons.put("tif",    "image.gif");
  icons.put("tiff",   "image.gif");
  icons.put("pnm",    "image.gif");
  icons.put("pbm",    "image.gif");
  icons.put("pgm",    "image.gif");
  icons.put("ppm",    "image.gif");
  icons.put("xbm",    "image.gif");
  icons.put("xpm",    "image.gif");
  icons.put("xwd",    "image.gif");
  icons.put("htm",    "text.gif");
  icons.put("html",   "text.gif");
  icons.put("xml",    "xml.gif");
  icons.put("xsl",    "xml.gif");
  icons.put("xslt",   "xml.gif");
  icons.put("xsd",    "xml.gif");
  icons.put("txt",    "text.gif");
  icons.put("text",   "text.gif");
  icons.put("mpg",    "video.gif");
  icons.put("mpe",    "video.gif");
  icons.put("mpeg",   "video.gif");
  icons.put("qt",     "video.gif");
  icons.put("mov",    "video.gif");
  icons.put("avi",    "video.gif");
  icons.put("movie",  "video.gif");
  icons.put("svg",    "image.gif");
  icons.put("css",    "text.gif");
  }
//...e
//...sinit:2:
public void init()
  throws ServletException
  {
  super.init();

  setupIcons();

  // See if init parameters are in a file
  // rather than in servlet initialisation parameters
  String propsfn;
  if ( (propsfn = getServletConfig().getInitParameter("webserver.propsfn")) != null )
    try
      {
      FileInputStream fis = new FileInputStream(propsfn);
      props = new Properties();
      props.load(fis);
      }
    catch ( IOException e )
      {
      throw new ServletException(
        "Can't load properties from "+propsfn+": "+e.getMessage());
      }
  else
    props = new Properties();

  // Determine all locations
  Vector<String> l = getInitParameterVector("locations");
  for ( int i = 0; i < l.size(); i++ )
    {
    String loc = l.elementAt(i);
    locations.add(new Location(loc));
    }

  // Global properties
  serverName = getInitParameter("name", "WebServer");
  logfile    = getInitParameter("logfile");
  }
//...e
  }
